package com.sun.org.apache.xalan.internal.xsltc.compiler;

import com.sun.org.apache.bcel.internal.generic.Instruction;
import com.sun.org.apache.bcel.internal.generic.DUP;
import com.sun.org.apache.bcel.internal.generic.ALOAD;
import com.sun.org.apache.bcel.internal.generic.TargetLostException;
import com.sun.org.apache.bcel.internal.util.InstructionFinder;
import com.sun.org.apache.bcel.internal.generic.BranchHandle;
import com.sun.org.apache.bcel.internal.generic.LocalVariableGen;
import com.sun.org.apache.bcel.internal.generic.IFLT;
import com.sun.org.apache.bcel.internal.generic.ISTORE;
import com.sun.org.apache.bcel.internal.generic.SWITCH;
import com.sun.org.apache.bcel.internal.generic.INVOKEVIRTUAL;
import com.sun.org.apache.bcel.internal.generic.INVOKEINTERFACE;
import com.sun.org.apache.bcel.internal.generic.ILOAD;
import java.util.Iterator;
import java.util.Set;
import com.sun.org.apache.bcel.internal.generic.GOTO_W;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.MethodGenerator;
import com.sun.org.apache.bcel.internal.generic.ConstantPoolGen;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.NamedMethodGenerator;
import com.sun.org.apache.bcel.internal.generic.Type;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.Util;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.ClassGenerator;
import java.util.Enumeration;
import java.util.HashMap;
import com.sun.org.apache.bcel.internal.generic.InstructionList;
import com.sun.org.apache.bcel.internal.generic.InstructionHandle;
import java.util.Map;
import java.util.Vector;

final class Mode implements Constants
{
    private final QName _name;
    private final Stylesheet _stylesheet;
    private final String _methodName;
    private Vector _templates;
    private Vector _childNodeGroup;
    private TestSeq _childNodeTestSeq;
    private Vector _attribNodeGroup;
    private TestSeq _attribNodeTestSeq;
    private Vector _idxGroup;
    private TestSeq _idxTestSeq;
    private Vector[] _patternGroups;
    private TestSeq[] _testSeq;
    private Map<Template, Object> _neededTemplates;
    private Map<Template, Mode> _namedTemplates;
    private Map<Template, InstructionHandle> _templateIHs;
    private Map<Template, InstructionList> _templateILs;
    private LocationPathPattern _rootPattern;
    private Map<Integer, Integer> _importLevels;
    private Map<String, Key> _keys;
    private int _currentIndex;
    
    public Mode(QName name, Stylesheet stylesheet, String suffix) {
        this._childNodeGroup = null;
        this._childNodeTestSeq = null;
        this._attribNodeGroup = null;
        this._attribNodeTestSeq = null;
        this._idxGroup = null;
        this._idxTestSeq = null;
        this._neededTemplates = new HashMap();
        this._namedTemplates = new HashMap();
        this._templateIHs = new HashMap();
        this._templateILs = new HashMap();
        this._rootPattern = null;
        this._importLevels = null;
        this._keys = null;
        this._name = name;
        this._stylesheet = stylesheet;
        this._methodName = "applyTemplates" + suffix;
        this._templates = new Vector();
        this._patternGroups = new Vector[32];
    }
    
    public String functionName() {
        return this._methodName;
    }
    
    public String functionName(int min, int max) {
        if (this._importLevels == null) {
            this._importLevels = new HashMap();
        }
        this._importLevels.put(Integer.valueOf(max), Integer.valueOf(min));
        return this._methodName + 95 + max;
    }
    
    private String getClassName() {
        return this._stylesheet.getClassName();
    }
    
    public Stylesheet getStylesheet() {
        return this._stylesheet;
    }
    
    public void addTemplate(Template template) {
        this._templates.addElement(template);
    }
    
    private Vector quicksort(Vector templates, int p, int r) {
        if (p < r) {
            int q = this.partition(templates, p, r);
            this.quicksort(templates, p, q);
            this.quicksort(templates, q + 1, r);
            return templates;
        }
        return templates;
    }
    
    private int partition(Vector templates, int p, int r) {
        Template x = (Template)templates.elementAt(p);
        int i = p - 1;
        int j = r + 1;
        while (true) {
            if (x.compareTo((Object)templates.elementAt(--j)) > 0) {
                continue;
            }
            while (x.compareTo((Object)templates.elementAt(++i)) < 0) {}
            if (i < j) {
                templates.set(j, templates.set(i, templates.elementAt(j)));
            }
            else {
                break;
            }
        }
        return j;
    }
    
    public void processPatterns(Map<String, Key> keys) {
        this._keys = keys;
        this._templates = this.quicksort(this._templates, 0, this._templates.size() - 1);
        Enumeration<Object> templates = this._templates.elements();
        while (templates.hasMoreElements()) {
            Template template = (Template)templates.nextElement();
            if (template.isNamed() && !template.disabled()) {
                this._namedTemplates.put(template, this);
            }
            Pattern pattern = template.getPattern();
            if (pattern != null) {
                this.flattenAlternative(pattern, template, keys);
            }
            else {
                continue;
            }
        }
        this.prepareTestSequences();
    }
    
    private void flattenAlternative(Pattern pattern, Template template, Map<String, Key> keys) {
        if ((pattern instanceof IdKeyPattern)) {
            IdKeyPattern idkey = (IdKeyPattern)pattern;
            idkey.setTemplate(template);
            if (this._idxGroup == null) {
                this._idxGroup = new Vector();
            }
            this._idxGroup.add(pattern);
        }
        else if ((pattern instanceof AlternativePattern)) {
            AlternativePattern alt = (AlternativePattern)pattern;
            this.flattenAlternative(alt.getLeft(), template, keys);
            this.flattenAlternative(alt.getRight(), template, keys);
        }
        else if ((pattern instanceof LocationPathPattern)) {
            LocationPathPattern lpp = (LocationPathPattern)pattern;
            lpp.setTemplate(template);
            this.addPatternToGroup(lpp);
        }
    }
    
    private void addPatternToGroup(LocationPathPattern lpp) {
        if ((lpp instanceof IdKeyPattern)) {
            this.addPattern(-1, lpp);
        }
        else {
            StepPattern kernel = lpp.getKernelPattern();
            if (kernel != null) {
                this.addPattern(kernel.getNodeType(), lpp);
            }
            else if (this._rootPattern == null || lpp.noSmallerThan(this._rootPattern)) {
                this._rootPattern = lpp;
            }
        }
    }
    
    private void addPattern(int kernelType, LocationPathPattern pattern) {
        int oldLength = this._patternGroups.length;
        if (kernelType >= oldLength) {
            Vector[] newGroups = new Vector[kernelType * 2];
            System.arraycopy(this._patternGroups, 0, newGroups, 0, oldLength);
            this._patternGroups = newGroups;
        }
        Vector<LocationPathPattern> patterns;
        if (kernelType == -1) {
            if (pattern.getAxis() == 2) {
                patterns = ((this._attribNodeGroup == null) ? (this._attribNodeGroup = new Vector(2)) : this._attribNodeGroup);
            }
            else {
                patterns = ((this._childNodeGroup == null) ? (this._childNodeGroup = new Vector(2)) : this._childNodeGroup);
            }
        }
        else {
            patterns = ((this._patternGroups[kernelType] == null) ? (this._patternGroups[kernelType] = new Vector(2)) : this._patternGroups[kernelType]);
        }
        if (patterns.size() == 0) {
            patterns.addElement(pattern);
        }
        else {
            boolean inserted = false;
            int i = 0;
            while (i < patterns.size()) {
                LocationPathPattern lppToCompare = (LocationPathPattern)patterns.elementAt(i);
                if (pattern.noSmallerThan(lppToCompare)) {
                    inserted = true;
                    patterns.insertElementAt(pattern, i);
                    break;
                }
                ++i;
            }
            if (!inserted) {
                patterns.addElement(pattern);
            }
        }
    }
    
    private void completeTestSequences(int nodeType, Vector patterns) {
        if (patterns != null) {
            if (this._patternGroups[nodeType] == null) {
                this._patternGroups[nodeType] = patterns;
            }
            else {
                for (int m = patterns.size(), j = 0; j < m; ++j) {
                    this.addPattern(nodeType, (LocationPathPattern)patterns.elementAt(j));
                }
            }
        }
    }
    
    private void prepareTestSequences() {
        Vector<Object> starGroup = this._patternGroups[1];
        Vector<Object> atStarGroup = this._patternGroups[2];
        this.completeTestSequences(3, this._childNodeGroup);
        this.completeTestSequences(1, this._childNodeGroup);
        this.completeTestSequences(7, this._childNodeGroup);
        this.completeTestSequences(8, this._childNodeGroup);
        this.completeTestSequences(2, this._attribNodeGroup);
        Vector<Object> names = this._stylesheet.getXSLTC().getNamesIndex();
        if (starGroup != null || atStarGroup != null || this._childNodeGroup != null || this._attribNodeGroup != null) {
            for (int n = this._patternGroups.length, i = 14; i < n; ++i) {
                if (this._patternGroups[i] != null) {
                    String name = (String)names.elementAt(i - 14);
                    if (isAttributeName(name)) {
                        this.completeTestSequences(i, atStarGroup);
                        this.completeTestSequences(i, this._attribNodeGroup);
                    }
                    else {
                        this.completeTestSequences(i, starGroup);
                        this.completeTestSequences(i, this._childNodeGroup);
                    }
                }
            }
        }
        this._testSeq = new TestSeq[14 + names.size()];
        for (int n = this._patternGroups.length, i = 0; i < n; ++i) {
            Vector<Object> patterns = this._patternGroups[i];
            if (patterns != null) {
                TestSeq testSeq = new TestSeq(patterns, i, this);
                testSeq.reduce();
                (this._testSeq[i] = testSeq).findTemplates(this._neededTemplates);
            }
        }
        if (this._childNodeGroup != null && this._childNodeGroup.size() > 0) {
            (this._childNodeTestSeq = new TestSeq(this._childNodeGroup, -1, this)).reduce();
            this._childNodeTestSeq.findTemplates(this._neededTemplates);
        }
        if (this._idxGroup != null && this._idxGroup.size() > 0) {
            (this._idxTestSeq = new TestSeq(this._idxGroup, this)).reduce();
            this._idxTestSeq.findTemplates(this._neededTemplates);
        }
        if (this._rootPattern != null) {
            this._neededTemplates.put(this._rootPattern.getTemplate(), this);
        }
    }
    
    private void compileNamedTemplate(Template template, ClassGenerator classGen) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = new InstructionList();
        String methodName = Util.escape(template.getName().toString());
        int numParams = 0;
        if (template.isSimpleNamedTemplate()) {
            Vector<Param> parameters = template.getParameters();
            numParams = parameters.size();
        }
        Type[] types = new Type[4 + numParams];
        String[] names = new String[4 + numParams];
        types[0] = Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;");
        types[1] = Util.getJCRefType("Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;");
        types[2] = Util.getJCRefType("Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;");
        types[3] = Type.INT;
        names[0] = "document";
        names[1] = "iterator";
        names[2] = "handler";
        names[3] = "node";
        for (int i = 4; i < 4 + numParams; ++i) {
            types[i] = Util.getJCRefType("Ljava/lang/Object;");
            names[i] = "param" + String.valueOf(i - 4);
        }
        NamedMethodGenerator methodGen = new NamedMethodGenerator(1, Type.VOID, types, names, methodName, this.getClassName(), il, cpg);
        il.append(template.compile(classGen, methodGen));
        il.append(Mode.RETURN);
        classGen.addMethod(methodGen);
    }
    
    private void compileTemplates(ClassGenerator classGen, MethodGenerator methodGen, InstructionHandle next) {
        Set<Template> templates = this._namedTemplates.keySet();
        for (Template template : templates) {
            this.compileNamedTemplate(template, classGen);
        }
        templates = this._neededTemplates.keySet();
        for (Template template : templates) {
            if (template.hasContents()) {
                InstructionList til = template.compile(classGen, methodGen);
                til.append(new GOTO_W(next));
                this._templateILs.put(template, til);
                this._templateIHs.put(template, til.getStart());
            }
            else {
                this._templateIHs.put(template, next);
            }
        }
    }
    
    private void appendTemplateCode(InstructionList body) {
        for (Template template : this._neededTemplates.keySet()) {
            InstructionList iList = (InstructionList)this._templateILs.get(template);
            if (iList != null) {
                body.append(iList);
            }
            else {
                continue;
            }
        }
    }
    
    private void appendTestSequences(InstructionList body) {
        for (int n = this._testSeq.length, i = 0; i < n; ++i) {
            TestSeq testSeq = this._testSeq[i];
            if (testSeq != null) {
                InstructionList il = testSeq.getInstructionList();
                if (il != null) {
                    body.append(il);
                }
            }
        }
    }
    
    public static void compileGetChildren(ClassGenerator classGen, MethodGenerator methodGen, int node) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = methodGen.getInstructionList();
        int git = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getChildren", "(I)Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;");
        il.append(methodGen.loadDOM());
        il.append(new ILOAD(node));
        il.append(new INVOKEINTERFACE(git, 2));
    }
    
    private InstructionList compileDefaultRecursion(ClassGenerator classGen, MethodGenerator methodGen, InstructionHandle next) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = new InstructionList();
        String applyTemplatesSig = classGen.getApplyTemplatesSig();
        int git = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getChildren", "(I)Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;");
        int applyTemplates = cpg.addMethodref(this.getClassName(), this.functionName(), applyTemplatesSig);
        il.append(classGen.loadTranslet());
        il.append(methodGen.loadDOM());
        il.append(methodGen.loadDOM());
        il.append(new ILOAD(this._currentIndex));
        il.append(new INVOKEINTERFACE(git, 2));
        il.append(methodGen.loadHandler());
        il.append(new INVOKEVIRTUAL(applyTemplates));
        il.append(new GOTO_W(next));
        return il;
    }
    
    private InstructionList compileDefaultText(ClassGenerator classGen, MethodGenerator methodGen, InstructionHandle next) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = new InstructionList();
        int chars = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "characters", "(ILcom/sun/org/apache/xml/internal/serializer/SerializationHandler;)V");
        il.append(methodGen.loadDOM());
        il.append(new ILOAD(this._currentIndex));
        il.append(methodGen.loadHandler());
        il.append(new INVOKEINTERFACE(chars, 3));
        il.append(new GOTO_W(next));
        return il;
    }
    
    private InstructionList compileNamespaces(ClassGenerator classGen, MethodGenerator methodGen, boolean[] isNamespace, boolean[] isAttribute, boolean attrFlag, InstructionHandle defaultTarget) {
        XSLTC xsltc = classGen.getParser().getXSLTC();
        ConstantPoolGen cpg = classGen.getConstantPool();
        Vector<Object> namespaces = xsltc.getNamespaceIndex();
        Vector<Object> names = xsltc.getNamesIndex();
        int namespaceCount = namespaces.size() + 1;
        int namesCount = names.size();
        InstructionList il = new InstructionList();
        int[] types = new int[namespaceCount];
        InstructionHandle[] targets = new InstructionHandle[types.length];
        if (namespaceCount > 0) {
            boolean compiled = false;
            for (int i = 0; i < namespaceCount; ++i) {
                targets[i] = defaultTarget;
                types[i] = i;
            }
            for (int i = 14; i < 14 + namesCount; ++i) {
                if (isNamespace[i] && isAttribute[i] == attrFlag) {
                    String name = (String)names.elementAt(i - 14);
                    String namespace = name.substring(0, name.lastIndexOf(58));
                    int type = xsltc.registerNamespace(namespace);
                    if (i < this._testSeq.length && this._testSeq[i] != null) {
                        targets[type] = this._testSeq[i].compile(classGen, methodGen, defaultTarget);
                        compiled = true;
                    }
                }
            }
            if (!compiled) {
                return null;
            }
            int getNS = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getNamespaceType", "(I)I");
            il.append(methodGen.loadDOM());
            il.append(new ILOAD(this._currentIndex));
            il.append(new INVOKEINTERFACE(getNS, 2));
            il.append(new SWITCH(types, targets, defaultTarget));
            return il;
        }
        else {
            return null;
        }
    }
    
    public void compileApplyTemplates(ClassGenerator classGen) {
        XSLTC xsltc = classGen.getParser().getXSLTC();
        ConstantPoolGen cpg = classGen.getConstantPool();
        Vector<Object> names = xsltc.getNamesIndex();
        Type[] argTypes = { Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;") };
        String[] argNames = { "document", "iterator", "handler" };
        InstructionList mainIL = new InstructionList();
        MethodGenerator methodGen = new MethodGenerator(17, Type.VOID, argTypes, argNames, this.functionName(), this.getClassName(), mainIL, classGen.getConstantPool());
        methodGen.addException("com.sun.org.apache.xalan.internal.xsltc.TransletException");
        mainIL.append(Mode.NOP);
        LocalVariableGen current = methodGen.addLocalVariable2("current", Type.INT, null);
        this._currentIndex = current.getIndex();
        InstructionList body = new InstructionList();
        body.append(Mode.NOP);
        InstructionList ilLoop = new InstructionList();
        ilLoop.append(methodGen.loadIterator());
        ilLoop.append(methodGen.nextNode());
        ilLoop.append(Mode.DUP);
        ilLoop.append(new ISTORE(this._currentIndex));
        BranchHandle ifeq = ilLoop.append(new IFLT(null));
        BranchHandle loop = ilLoop.append(new GOTO_W(null));
        ifeq.setTarget(ilLoop.append(Mode.RETURN));
        InstructionHandle ihLoop = ilLoop.getStart();
        current.setStart(mainIL.append(new GOTO_W(ihLoop)));
        current.setEnd(loop);
        InstructionList ilRecurse = this.compileDefaultRecursion(classGen, methodGen, ihLoop);
        InstructionHandle ihRecurse = ilRecurse.getStart();
        InstructionList ilText = this.compileDefaultText(classGen, methodGen, ihLoop);
        InstructionHandle ihText = ilText.getStart();
        int[] types = new int[14 + names.size()];
        for (int i = 0; i < types.length; ++i) {
            types[i] = i;
        }
        boolean[] isAttribute = new boolean[types.length];
        boolean[] isNamespace = new boolean[types.length];
        for (int j = 0; j < names.size(); ++j) {
            String name = (String)names.elementAt(j);
            isAttribute[j + 14] = isAttributeName(name);
            isNamespace[j + 14] = isNamespaceName(name);
        }
        this.compileTemplates(classGen, methodGen, ihLoop);
        TestSeq elemTest = this._testSeq[1];
        InstructionHandle ihElem = ihRecurse;
        if (elemTest != null) {
            ihElem = elemTest.compile(classGen, methodGen, ihRecurse);
        }
        TestSeq attrTest = this._testSeq[2];
        InstructionHandle ihAttr = ihText;
        if (attrTest != null) {
            ihAttr = attrTest.compile(classGen, methodGen, ihAttr);
        }
        InstructionList ilKey = null;
        if (this._idxTestSeq != null) {
            loop.setTarget(this._idxTestSeq.compile(classGen, methodGen, body.getStart()));
            ilKey = this._idxTestSeq.getInstructionList();
        }
        else {
            loop.setTarget(body.getStart());
        }
        if (this._childNodeTestSeq != null) {
            double nodePrio = this._childNodeTestSeq.getPriority();
            int nodePos = this._childNodeTestSeq.getPosition();
            double elemPrio = -1.7976931348623157E308;
            int elemPos = Integer.MIN_VALUE;
            if (elemTest != null) {
                elemPrio = elemTest.getPriority();
                elemPos = elemTest.getPosition();
            }
            if (elemPrio == Double.NaN || elemPrio < nodePrio || (elemPrio == nodePrio && elemPos < nodePos)) {
                ihElem = this._childNodeTestSeq.compile(classGen, methodGen, ihLoop);
            }
            TestSeq textTest = this._testSeq[3];
            double textPrio = -1.7976931348623157E308;
            int textPos = Integer.MIN_VALUE;
            if (textTest != null) {
                textPrio = textTest.getPriority();
                textPos = textTest.getPosition();
            }
            if (textPrio == Double.NaN || textPrio < nodePrio || (textPrio == nodePrio && textPos < nodePos)) {
                ihText = this._childNodeTestSeq.compile(classGen, methodGen, ihLoop);
                this._testSeq[3] = this._childNodeTestSeq;
            }
        }
        InstructionHandle elemNamespaceHandle = ihElem;
        InstructionList nsElem = this.compileNamespaces(classGen, methodGen, isNamespace, isAttribute, false, ihElem);
        if (nsElem != null) {
            elemNamespaceHandle = nsElem.getStart();
        }
        InstructionHandle attrNamespaceHandle = ihAttr;
        InstructionList nsAttr = this.compileNamespaces(classGen, methodGen, isNamespace, isAttribute, true, ihAttr);
        if (nsAttr != null) {
            attrNamespaceHandle = nsAttr.getStart();
        }
        InstructionHandle[] targets = new InstructionHandle[types.length];
        for (int k = 14; k < targets.length; ++k) {
            TestSeq testSeq = this._testSeq[k];
            if (isNamespace[k]) {
                if (isAttribute[k]) {
                    targets[k] = attrNamespaceHandle;
                }
                else {
                    targets[k] = elemNamespaceHandle;
                }
            }
            else if (testSeq != null) {
                if (isAttribute[k]) {
                    targets[k] = testSeq.compile(classGen, methodGen, attrNamespaceHandle);
                }
                else {
                    targets[k] = testSeq.compile(classGen, methodGen, elemNamespaceHandle);
                }
            }
            else {
                targets[k] = ihLoop;
            }
        }
        targets[0] = ((this._rootPattern != null) ? this.getTemplateInstructionHandle(this._rootPattern.getTemplate()) : ihRecurse);
        targets[9] = ((this._rootPattern != null) ? this.getTemplateInstructionHandle(this._rootPattern.getTemplate()) : ihRecurse);
        targets[3] = ((this._testSeq[3] != null) ? this._testSeq[3].compile(classGen, methodGen, ihText) : ihText);
        targets[13] = ihLoop;
        targets[1] = elemNamespaceHandle;
        targets[2] = attrNamespaceHandle;
        InstructionHandle ihPI = ihLoop;
        if (this._childNodeTestSeq != null) {
            ihPI = ihElem;
        }
        if (this._testSeq[7] != null) {
            targets[7] = this._testSeq[7].compile(classGen, methodGen, ihPI);
        }
        else {
            targets[7] = ihPI;
        }
        InstructionHandle ihComment = ihLoop;
        if (this._childNodeTestSeq != null) {
            ihComment = ihElem;
        }
        targets[8] = ((this._testSeq[8] != null) ? this._testSeq[8].compile(classGen, methodGen, ihComment) : ihComment);
        targets[11] = (targets[4] = ihLoop);
        targets[6] = (targets[10] = ihLoop);
        targets[12] = (targets[5] = ihLoop);
        for (int l = 14; l < targets.length; ++l) {
            TestSeq testSeq2 = this._testSeq[l];
            if (testSeq2 == null || isNamespace[l]) {
                if (isAttribute[l]) {
                    targets[l] = attrNamespaceHandle;
                }
                else {
                    targets[l] = elemNamespaceHandle;
                }
            }
            else if (isAttribute[l]) {
                targets[l] = testSeq2.compile(classGen, methodGen, attrNamespaceHandle);
            }
            else {
                targets[l] = testSeq2.compile(classGen, methodGen, elemNamespaceHandle);
            }
        }
        if (ilKey != null) {
            body.insert(ilKey);
        }
        int getType = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getExpandedTypeID", "(I)I");
        body.append(methodGen.loadDOM());
        body.append(new ILOAD(this._currentIndex));
        body.append(new INVOKEINTERFACE(getType, 2));
        InstructionHandle disp = body.append(new SWITCH(types, targets, ihLoop));
        this.appendTestSequences(body);
        this.appendTemplateCode(body);
        if (nsElem != null) {
            body.append(nsElem);
        }
        if (nsAttr != null) {
            body.append(nsAttr);
        }
        body.append(ilRecurse);
        body.append(ilText);
        mainIL.append(body);
        mainIL.append(ilLoop);
        this.peepHoleOptimization(methodGen);
        classGen.addMethod(methodGen);
        if (this._importLevels != null) {
            for (Map.Entry<Integer, Integer> entry : this._importLevels.entrySet()) {
                this.compileApplyImports(classGen, (int)entry.getValue(), (int)entry.getKey());
            }
        }
    }
    
    private void compileTemplateCalls(ClassGenerator classGen, MethodGenerator methodGen, InstructionHandle next, int min, int max) {
        for (Template template : this._neededTemplates.keySet()) {
            int prec = template.getImportPrecedence();
            if (prec >= min && prec < max) {
                if (template.hasContents()) {
                    InstructionList til = template.compile(classGen, methodGen);
                    til.append(new GOTO_W(next));
                    this._templateILs.put(template, til);
                    this._templateIHs.put(template, til.getStart());
                }
                else {
                    this._templateIHs.put(template, next);
                }
            }
            else {
                continue;
            }
        }
    }
    
    public void compileApplyImports(ClassGenerator classGen, int min, int max) {
        XSLTC xsltc = classGen.getParser().getXSLTC();
        ConstantPoolGen cpg = classGen.getConstantPool();
        Vector<Object> names = xsltc.getNamesIndex();
        this._namedTemplates = new HashMap();
        this._neededTemplates = new HashMap();
        this._templateIHs = new HashMap();
        this._templateILs = new HashMap();
        this._patternGroups = new Vector[32];
        this._rootPattern = null;
        Vector<Object> oldTemplates = this._templates;
        this._templates = new Vector();
        Enumeration<Object> templates = oldTemplates.elements();
        while (templates.hasMoreElements()) {
            Template template = (Template)templates.nextElement();
            int prec = template.getImportPrecedence();
            if (prec >= min && prec < max) {
                this.addTemplate(template);
            }
            else {
                continue;
            }
        }
        this.processPatterns(this._keys);
        Type[] argTypes = { Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;"), Type.INT };
        String[] argNames = { "document", "iterator", "handler", "node" };
        InstructionList mainIL = new InstructionList();
        MethodGenerator methodGen = new MethodGenerator(17, Type.VOID, argTypes, argNames, this.functionName() + 95 + max, this.getClassName(), mainIL, classGen.getConstantPool());
        methodGen.addException("com.sun.org.apache.xalan.internal.xsltc.TransletException");
        LocalVariableGen current = methodGen.addLocalVariable2("current", Type.INT, null);
        this._currentIndex = current.getIndex();
        mainIL.append(new ILOAD(methodGen.getLocalIndex("node")));
        current.setStart(mainIL.append(new ISTORE(this._currentIndex)));
        InstructionList body = new InstructionList();
        body.append(Mode.NOP);
        InstructionList ilLoop = new InstructionList();
        ilLoop.append(Mode.RETURN);
        InstructionHandle ihLoop = ilLoop.getStart();
        InstructionList ilRecurse = this.compileDefaultRecursion(classGen, methodGen, ihLoop);
        InstructionHandle ihRecurse = ilRecurse.getStart();
        InstructionList ilText = this.compileDefaultText(classGen, methodGen, ihLoop);
        InstructionHandle ihText = ilText.getStart();
        int[] types = new int[14 + names.size()];
        for (int i = 0; i < types.length; ++i) {
            types[i] = i;
        }
        boolean[] isAttribute = new boolean[types.length];
        boolean[] isNamespace = new boolean[types.length];
        for (int j = 0; j < names.size(); ++j) {
            String name = (String)names.elementAt(j);
            isAttribute[j + 14] = isAttributeName(name);
            isNamespace[j + 14] = isNamespaceName(name);
        }
        this.compileTemplateCalls(classGen, methodGen, ihLoop, min, max);
        TestSeq elemTest = this._testSeq[1];
        InstructionHandle ihElem = ihRecurse;
        if (elemTest != null) {
            ihElem = elemTest.compile(classGen, methodGen, ihLoop);
        }
        TestSeq attrTest = this._testSeq[2];
        InstructionHandle ihAttr = ihLoop;
        if (attrTest != null) {
            ihAttr = attrTest.compile(classGen, methodGen, ihAttr);
        }
        InstructionList ilKey = null;
        if (this._idxTestSeq != null) {
            ilKey = this._idxTestSeq.getInstructionList();
        }
        if (this._childNodeTestSeq != null) {
            double nodePrio = this._childNodeTestSeq.getPriority();
            int nodePos = this._childNodeTestSeq.getPosition();
            double elemPrio = -1.7976931348623157E308;
            int elemPos = Integer.MIN_VALUE;
            if (elemTest != null) {
                elemPrio = elemTest.getPriority();
                elemPos = elemTest.getPosition();
            }
            if (elemPrio == Double.NaN || elemPrio < nodePrio || (elemPrio == nodePrio && elemPos < nodePos)) {
                ihElem = this._childNodeTestSeq.compile(classGen, methodGen, ihLoop);
            }
            TestSeq textTest = this._testSeq[3];
            double textPrio = -1.7976931348623157E308;
            int textPos = Integer.MIN_VALUE;
            if (textTest != null) {
                textPrio = textTest.getPriority();
                textPos = textTest.getPosition();
            }
            if (textPrio == Double.NaN || textPrio < nodePrio || (textPrio == nodePrio && textPos < nodePos)) {
                ihText = this._childNodeTestSeq.compile(classGen, methodGen, ihLoop);
                this._testSeq[3] = this._childNodeTestSeq;
            }
        }
        InstructionHandle elemNamespaceHandle = ihElem;
        InstructionList nsElem = this.compileNamespaces(classGen, methodGen, isNamespace, isAttribute, false, ihElem);
        if (nsElem != null) {
            elemNamespaceHandle = nsElem.getStart();
        }
        InstructionList nsAttr = this.compileNamespaces(classGen, methodGen, isNamespace, isAttribute, true, ihAttr);
        InstructionHandle attrNamespaceHandle = ihAttr;
        if (nsAttr != null) {
            attrNamespaceHandle = nsAttr.getStart();
        }
        InstructionHandle[] targets = new InstructionHandle[types.length];
        for (int k = 14; k < targets.length; ++k) {
            TestSeq testSeq = this._testSeq[k];
            if (isNamespace[k]) {
                if (isAttribute[k]) {
                    targets[k] = attrNamespaceHandle;
                }
                else {
                    targets[k] = elemNamespaceHandle;
                }
            }
            else if (testSeq != null) {
                if (isAttribute[k]) {
                    targets[k] = testSeq.compile(classGen, methodGen, attrNamespaceHandle);
                }
                else {
                    targets[k] = testSeq.compile(classGen, methodGen, elemNamespaceHandle);
                }
            }
            else {
                targets[k] = ihLoop;
            }
        }
        targets[0] = ((this._rootPattern != null) ? this.getTemplateInstructionHandle(this._rootPattern.getTemplate()) : ihRecurse);
        targets[9] = ((this._rootPattern != null) ? this.getTemplateInstructionHandle(this._rootPattern.getTemplate()) : ihRecurse);
        targets[3] = ((this._testSeq[3] != null) ? this._testSeq[3].compile(classGen, methodGen, ihText) : ihText);
        targets[13] = ihLoop;
        targets[1] = elemNamespaceHandle;
        targets[2] = attrNamespaceHandle;
        InstructionHandle ihPI = ihLoop;
        if (this._childNodeTestSeq != null) {
            ihPI = ihElem;
        }
        if (this._testSeq[7] != null) {
            targets[7] = this._testSeq[7].compile(classGen, methodGen, ihPI);
        }
        else {
            targets[7] = ihPI;
        }
        InstructionHandle ihComment = ihLoop;
        if (this._childNodeTestSeq != null) {
            ihComment = ihElem;
        }
        targets[8] = ((this._testSeq[8] != null) ? this._testSeq[8].compile(classGen, methodGen, ihComment) : ihComment);
        targets[11] = (targets[4] = ihLoop);
        targets[6] = (targets[10] = ihLoop);
        targets[12] = (targets[5] = ihLoop);
        for (int l = 14; l < targets.length; ++l) {
            TestSeq testSeq2 = this._testSeq[l];
            if (testSeq2 == null || isNamespace[l]) {
                if (isAttribute[l]) {
                    targets[l] = attrNamespaceHandle;
                }
                else {
                    targets[l] = elemNamespaceHandle;
                }
            }
            else if (isAttribute[l]) {
                targets[l] = testSeq2.compile(classGen, methodGen, attrNamespaceHandle);
            }
            else {
                targets[l] = testSeq2.compile(classGen, methodGen, elemNamespaceHandle);
            }
        }
        if (ilKey != null) {
            body.insert(ilKey);
        }
        int getType = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getExpandedTypeID", "(I)I");
        body.append(methodGen.loadDOM());
        body.append(new ILOAD(this._currentIndex));
        body.append(new INVOKEINTERFACE(getType, 2));
        InstructionHandle disp = body.append(new SWITCH(types, targets, ihLoop));
        this.appendTestSequences(body);
        this.appendTemplateCode(body);
        if (nsElem != null) {
            body.append(nsElem);
        }
        if (nsAttr != null) {
            body.append(nsAttr);
        }
        body.append(ilRecurse);
        body.append(ilText);
        mainIL.append(body);
        current.setEnd(body.getEnd());
        mainIL.append(ilLoop);
        this.peepHoleOptimization(methodGen);
        classGen.addMethod(methodGen);
        this._templates = oldTemplates;
    }
    
    private void peepHoleOptimization(MethodGenerator methodGen) {
        InstructionList il = methodGen.getInstructionList();
        InstructionFinder find = new InstructionFinder(il);
        String pattern = "loadinstruction pop";
        Iterator<InstructionHandle> iter = find.search(pattern);
        while (iter.hasNext()) {
            InstructionHandle[] match = (InstructionHandle[])(InstructionHandle[])iter.next();
            try {
                if (!match[0].hasTargeters() && !match[1].hasTargeters()) {
                    il.delete(match[0], match[1]);
                }
                else {
                    continue;
                }
            }
            catch (TargetLostException ex) {}
        }
        pattern = "iload iload swap istore";
        iter = find.search(pattern);
        while (iter.hasNext()) {
            InstructionHandle[] match = (InstructionHandle[])(InstructionHandle[])iter.next();
            try {
                ILOAD iload1 = (ILOAD)match[0].getInstruction();
                ILOAD iload2 = (ILOAD)match[1].getInstruction();
                ISTORE istore = (ISTORE)match[3].getInstruction();
                if (!match[1].hasTargeters() && !match[2].hasTargeters() && !match[3].hasTargeters() && iload1.getIndex() == iload2.getIndex() && iload2.getIndex() == istore.getIndex()) {
                    il.delete(match[1], match[3]);
                }
                else {
                    continue;
                }
            }
            catch (TargetLostException ex2) {}
        }
        pattern = "loadinstruction loadinstruction swap";
        iter = find.search(pattern);
        while (iter.hasNext()) {
            InstructionHandle[] match = (InstructionHandle[])(InstructionHandle[])iter.next();
            try {
                if (!match[0].hasTargeters() && !match[1].hasTargeters() && !match[2].hasTargeters()) {
                    Instruction load_m = match[1].getInstruction();
                    il.insert(match[0], load_m);
                    il.delete(match[1], match[2]);
                }
                else {
                    continue;
                }
            }
            catch (TargetLostException ex3) {}
        }
        pattern = "aload aload";
        iter = find.search(pattern);
        while (iter.hasNext()) {
            InstructionHandle[] match = (InstructionHandle[])(InstructionHandle[])iter.next();
            try {
                if (!match[1].hasTargeters()) {
                    ALOAD aload1 = (ALOAD)match[0].getInstruction();
                    ALOAD aload2 = (ALOAD)match[1].getInstruction();
                    if (aload1.getIndex() == aload2.getIndex()) {
                        il.insert(match[1], new DUP());
                        il.delete(match[1]);
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
            catch (TargetLostException ex4) {}
        }
    }
    
    public InstructionHandle getTemplateInstructionHandle(Template template) {
        return (InstructionHandle)this._templateIHs.get(template);
    }
    
    private static boolean isAttributeName(String qname) {
        int col = qname.lastIndexOf(58) + 1;
        return qname.charAt(col) == '@';
    }
    
    private static boolean isNamespaceName(String qname) {
        int col = qname.lastIndexOf(58);
        return col > -1 && qname.charAt(qname.length() - 1) == '*';
    }
}
